import os
import shutil
import tempfile
import filecmp
import unittest.mock
import sqlalchemy.orm
import gffutils
from .. import utils
from ..io import iobase, fasta, gff
from ..orm import general, cv, organism, pub, sequence

modules_dir = os.path.dirname(os.path.abspath(gff.__file__))
data_dir = os.path.abspath(os.path.join(modules_dir, '..', 'tests', 'data'))


class TestGFFImport(unittest.TestCase):
    """Tests various functions used to load a GFF file into a database"""

    @classmethod
    def setUpClass(cls):
        # Creates an instance of the base class to be tested and instantiates global attributes
        cls.client = gff.GFFImportClient("testuri", test_environment=True)
        cls.client._synonym_terms = {
            "synonym": cv.CvTerm(cv_id=3, dbxref_id=31, name="synonym", cvterm_id=31),
            "previous_systematic_id": cv.CvTerm(cv_id=3, dbxref_id=32, name="previous_systematic_id", cvterm_id=32),
            "alias": cv.CvTerm(cv_id=3, dbxref_id=33, name="alias", cvterm_id=33)
        }
        cls.client._sequence_terms = {
            "gene": cv.CvTerm(cv_id=4, dbxref_id=41, name="gene", cvterm_id=41),
            "mRNA": cv.CvTerm(cv_id=4, dbxref_id=42, name="gene", cvterm_id=42),
            "chromosome": cv.CvTerm(cv_id=4, dbxref_id=43, name="chromosome", cvterm_id=43)
        }
        cls.client._feature_property_terms = {
            "score": cv.CvTerm(cv_id=5, dbxref_id=51, name="score", cvterm_id=51),
            "source": cv.CvTerm(cv_id=5, dbxref_id=52, name="source", cvterm_id=52),
            "comment": cv.CvTerm(cv_id=5, dbxref_id=53, name="comment", cvterm_id=53)
        }
        cls.client._parent_terms = {
            "part_of": cv.CvTerm(cv_id=6, dbxref_id=62, name="part_of", cvterm_id=62, is_relationshiptype=1),
            "derives_from": cv.CvTerm(cv_id=6, dbxref_id=63, name="derives_from", cvterm_id=63, is_relationshiptype=1)
        }
        cls.client._default_pub = pub.Pub(uniquename="defaultpub", type_id=71, pub_id=33)
        cls.client._synonym_type_ids = [31, 32, 33]
        cls.client._feature_property_type_ids = [51, 52, 53]
        cls.client._parent_type_ids = [62, 63]
        cls.client._go_db = general.Db(db_id=131, name="GO")

    def setUp(self):
        # Creates a default GFF record
        self.default_gff_record = gffutils.Feature(
            id="testid", seqid="testseqid", source="testsource", featuretype="testtype", start=1, end=30, score="3.5",
            strand="+", frame="2", attributes={
                "Name": ["testname"], "translation": ["MCRA"], "literature": ["PMID:12334"], "Alias": ["testalias"],
                "previous_systematic_id": "testsynonym", "Parent": "testparent", "Dbxref": ["testdb:testaccession"],
                "Ontology_term": ["GO:7890"], "Note": "testnote"})

    def test_create_sqlite_db(self):
        # Tests the function that creates a SQLite database from a GFF file
        gff_file = os.path.join(data_dir, 'gff_without_fasta.gff3')
        self.assertTrue(os.path.exists(gff_file))
        gff_db = self.client._create_sqlite_db(gff_file)
        self.assertEqual(len(self.client._sqlite_databases), 1)
        gff_db_name = self.client._sqlite_databases[0]
        self.assertTrue(os.path.exists(gff_db_name))
        self.assertIn("gff-version 3", gff_db.directives)
        os.remove(gff_db_name)
        self.assertFalse(os.path.exists(gff_db_name))

    def test_has_fasta(self):
        # Tests the function that checks if a GFF file contains a FASTA section
        gff_file = os.path.join(data_dir, 'gff_without_fasta.gff3')
        self.assertFalse(self.client._has_fasta(gff_file))
        gff_file = os.path.join(data_dir, 'gff_with_fasta.gff3')
        self.assertTrue(self.client._has_fasta(gff_file))

    def test_split_off_fasta(self):
        # Tests the function copying the FASTA section from a GFF file into a separate file
        gff_file = os.path.join(data_dir, 'gff_with_fasta.gff3')
        fasta_file = tempfile.mkstemp()[1]
        self.client._split_off_fasta(gff_file, fasta_file)
        actual_fasta_file = os.path.join(data_dir, 'fasta_only.fa')
        self.assertTrue(filecmp.cmp(fasta_file, actual_fasta_file))
        os.remove(fasta_file)

    @unittest.mock.patch("pychado.io.fasta.FastaImportClient")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient._split_off_fasta")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient._has_fasta")
    def test_import_fasta(self, mock_includes: unittest.mock.Mock, mock_copy: unittest.mock.Mock,
                          mock_fasta: unittest.mock.Mock):
        # Tests the function that imports the FASTA sequences for a GFF file
        self.assertIs(mock_includes, self.client._has_fasta)
        self.assertIs(mock_copy, self.client._split_off_fasta)
        self.assertIs(mock_fasta, fasta.FastaImportClient)

        # FASTA in GFF and separate file
        mock_includes.return_value = True
        with self.assertRaises(iobase.InputFileError):
            self.client._import_fasta("testgff", "testfasta", "testorganism", "region")

        # FASTA in separate file only
        mock_includes.return_value = False
        mock_copy.reset_mock()
        mock_fasta.reset_mock()
        self.client._import_fasta("testgff", "testfasta", "testorganism", "region")
        mock_includes.assert_called_with("testgff")
        mock_copy.assert_not_called()
        mock_fasta.assert_called_with("testuri", False)
        self.assertIn(unittest.mock.call().load("testfasta", "testorganism", "region"), mock_fasta.mock_calls)

        # FASTA in GFF only
        mock_includes.return_value = True
        mock_copy.reset_mock()
        mock_fasta.reset_mock()
        self.client._import_fasta("testgff", "", "testorganism", "region")
        mock_includes.assert_called_with("testgff")
        mock_copy.assert_called()
        mock_fasta.assert_called()

        # No FASTA
        mock_includes.return_value = False
        mock_copy.reset_mock()
        mock_fasta.reset_mock()
        self.client._import_fasta("testgff", "", "testorganism", "region")
        mock_includes.assert_called_with("testgff")
        mock_copy.assert_not_called()
        mock_fasta.assert_not_called()

    @unittest.mock.patch("pychado.io.gff.GFFImportClient.query_table")
    def test_handle_existing_features(self, mock_query: unittest.mock.Mock):
        self.assertIs(mock_query, self.client.query_table)
        organism_entry = organism.Organism(genus="", species="", abbreviation="testorganism", organism_id=1)
        mock_query.return_value = unittest.mock.Mock(sqlalchemy.orm.Query)

        # update import - should not query
        self.client._handle_existing_features(organism_entry, False, False)
        mock_query.assert_not_called()

        # from-scratch import on empty table - should not call delete
        mock_query_object = mock_query.return_value
        mock_query_object.configure_mock(**{"count.return_value": 0})
        self.client._handle_existing_features(organism_entry, True, False)
        mock_query.assert_called_with(sequence.Feature, organism_id=organism_entry.organism_id)
        self.assertIn(unittest.mock.call.count(), mock_query_object.method_calls)
        self.assertEqual(len(mock_query_object.mock_calls), 1)

        # from-scratch import on full table without 'force' - should throw an exception
        mock_query_object.reset_mock()
        mock_query_object.configure_mock(**{"count.return_value": 1})
        with self.assertRaises(iobase.DatabaseError):
            self.client._handle_existing_features(organism_entry, True, False)
        self.assertIn(unittest.mock.call.count(), mock_query_object.method_calls)
        self.assertEqual(len(mock_query_object.mock_calls), 1)

        # from-scratch import on full table with 'force' - should call delete
        mock_query_object.reset_mock()
        mock_query_object.configure_mock(**{"count.return_value": 1})
        self.client._handle_existing_features(organism_entry, True, True)
        self.assertIn(unittest.mock.call.count(), mock_query_object.method_calls)
        self.assertIn(unittest.mock.call.delete(), mock_query_object.method_calls)
        self.assertEqual(len(mock_query_object.mock_calls), 2)

    @unittest.mock.patch("pychado.io.gff.GFFImportClient._check_if_gff_attributes_are_recognized")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient._handle_protein")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient._handle_relationships")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient._handle_publications")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient._handle_ontology_terms")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient._handle_cross_references")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient._handle_properties")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient._handle_synonyms")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient._handle_location")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient._handle_child_feature")
    def test_insert_gff_record_into_database(
            self, mock_handle_child: unittest.mock.Mock, mock_handle_location: unittest.mock.Mock,
            mock_handle_synonyms: unittest.mock.Mock, mock_handle_properties: unittest.mock.Mock,
            mock_handle_cross_references: unittest.mock.Mock, mock_handle_ontology_terms: unittest.mock.Mock,
            mock_handle_publications: unittest.mock.Mock, mock_handle_relationships: unittest.mock.Mock,
            mock_handle_protein: unittest.mock.Mock, mock_check_recognized: unittest.mock.Mock):
        # Tests the main function updating database tables according to the information in a GFF record
        self.assertIs(mock_handle_child, self.client._handle_child_feature)
        self.assertIs(mock_handle_location, self.client._handle_location)
        self.assertIs(mock_handle_synonyms, self.client._handle_synonyms)
        self.assertIs(mock_handle_properties, self.client._handle_properties)
        self.assertIs(mock_handle_cross_references, self.client._handle_cross_references)
        self.assertIs(mock_handle_ontology_terms, self.client._handle_ontology_terms)
        self.assertIs(mock_handle_publications, self.client._handle_publications)
        self.assertIs(mock_handle_relationships, self.client._handle_relationships)
        self.assertIs(mock_handle_protein, self.client._handle_protein)
        self.assertIs(mock_check_recognized, self.client._check_if_gff_attributes_are_recognized)

        organism_entry = organism.Organism(genus="", species="", abbreviation="testorganism", organism_id=1)
        all_features = {}
        mock_handle_child.return_value = None
        self.client._insert_gff_record_into_database(self.default_gff_record, organism_entry, all_features)
        mock_handle_child.assert_called_with(self.default_gff_record, organism_entry)
        mock_handle_location.assert_not_called()
        self.assertEqual(len(all_features), 0)

        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testname", feature_id=1)
        mock_handle_child.return_value = feature_entry
        self.client._insert_gff_record_into_database(self.default_gff_record, organism_entry, all_features)
        mock_handle_child.assert_called_with(self.default_gff_record, organism_entry)
        mock_handle_location.assert_called_with(self.default_gff_record, feature_entry)
        mock_handle_synonyms.assert_called_with(self.default_gff_record, feature_entry)
        mock_handle_properties.assert_called_with(self.default_gff_record, feature_entry)
        mock_handle_cross_references.assert_called_with(self.default_gff_record, feature_entry)
        mock_handle_ontology_terms.assert_called_with(self.default_gff_record, feature_entry)
        mock_handle_publications.assert_called_with(self.default_gff_record, feature_entry)
        mock_handle_relationships.assert_called_with(self.default_gff_record, feature_entry, all_features)
        mock_handle_protein.assert_called_with(self.default_gff_record, feature_entry, organism_entry, all_features)
        mock_check_recognized.assert_called_with(self.default_gff_record)
        self.assertEqual(len(all_features), 1)

    @unittest.mock.patch("pychado.io.gff.GFFImportClient._handle_feature")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient._create_feature")
    def test_handle_child_feature(self, mock_create: unittest.mock.Mock, mock_insert: unittest.mock.Mock):
        # Tests the function transferring data from a GFF record to the 'feature' table
        self.assertIs(mock_create, self.client._create_feature)
        self.assertIs(mock_insert, self.client._handle_feature)
        organism_entry = organism.Organism(genus="", species="", abbreviation="testorganism", organism_id=1)
        feature_entry = self.client._handle_child_feature(self.default_gff_record, organism_entry)
        self.assertIsNone(feature_entry)
        mock_create.assert_not_called()

        self.default_gff_record.featuretype = "gene"
        mock_create.return_value = "AAA"
        self.client._handle_child_feature(self.default_gff_record, organism_entry)
        mock_create.assert_called_with(self.default_gff_record, 1, 41)
        mock_insert.assert_called_with("AAA", "testorganism")

    @unittest.mock.patch("pychado.io.gff.GFFImportClient._handle_featureloc")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient._create_featureloc")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient.query_first")
    def test_handle_location(self, mock_query: unittest.mock.Mock, mock_create: unittest.mock.Mock,
                             mock_insert: unittest.mock.Mock):
        # Tests the function transferring data from a GFF record to the 'feature' table
        self.assertIs(mock_query, self.client.query_first)
        self.assertIs(mock_create, self.client._create_featureloc)
        self.assertIs(mock_insert, self.client._handle_featureloc)

        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testname", feature_id=1)
        mock_query.return_value = sequence.Feature(organism_id=11, type_id=300, uniquename="chromname", feature_id=2)

        featureloc_entry = self.client._handle_location(self.default_gff_record, feature_entry)
        mock_query.assert_called_with(sequence.Feature, organism_id=11, uniquename="testseqid")
        mock_create.assert_called_with(self.default_gff_record, 1, 2)
        mock_insert.assert_called()
        self.assertIsNotNone(featureloc_entry)

        mock_query.return_value = None
        featureloc_entry = self.client._handle_location(self.default_gff_record, feature_entry)
        self.assertIsNone(featureloc_entry)

    @unittest.mock.patch("pychado.io.gff.GFFImportClient._delete_feature_synonym")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient._handle_feature_synonym")
    @unittest.mock.patch("pychado.orm.sequence.FeatureSynonym")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient._handle_synonym")
    @unittest.mock.patch("pychado.orm.sequence.Synonym")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient.query_feature_synonym_by_type")
    def test_handle_synonyms(self, mock_query: unittest.mock.Mock, mock_synonym: unittest.mock.Mock,
                             mock_insert_synonym: unittest.mock.Mock, mock_feature_synonym: unittest.mock.Mock,
                             mock_insert_feature_synonym: unittest.mock.Mock,
                             mock_delete_feature_synonym: unittest.mock.Mock):
        # Tests the function transferring data from a GFF record to the 'feature_synonym' table
        self.assertIs(mock_query, self.client.query_feature_synonym_by_type)
        self.assertIs(mock_synonym, sequence.Synonym)
        self.assertIs(mock_insert_synonym, self.client._handle_synonym)
        self.assertIs(mock_feature_synonym, sequence.FeatureSynonym)
        self.assertIs(mock_insert_feature_synonym, self.client._handle_feature_synonym)
        self.assertIs(mock_delete_feature_synonym, self.client._delete_feature_synonym)

        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testname", feature_id=1)
        mock_insert_synonym.return_value = utils.EmptyObject(synonym_id=12)

        all_synonyms = self.client._handle_synonyms(self.default_gff_record, feature_entry)
        mock_query.assert_called_with(1, [31, 32, 33])
        mock_synonym.assert_any_call(name="testalias", type_id=33, synonym_sgml="testalias")
        self.assertEqual(mock_insert_synonym.call_count, 2)
        mock_feature_synonym.assert_any_call(synonym_id=12, feature_id=1, pub_id=33)
        self.assertEqual(mock_insert_feature_synonym.call_count, 2)
        mock_delete_feature_synonym.assert_called()
        self.assertEqual(len(all_synonyms), 2)

    @unittest.mock.patch("pychado.io.gff.GFFImportClient._delete_feature_pub")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient._handle_feature_pub")
    @unittest.mock.patch("pychado.orm.sequence.FeaturePub")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient._handle_pub")
    @unittest.mock.patch("pychado.orm.pub.Pub")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient.query_all")
    def test_handle_publications(self, mock_query: unittest.mock.Mock, mock_pub: unittest.mock.Mock,
                                 mock_insert_pub: unittest.mock.Mock, mock_featurepub: unittest.mock.Mock,
                                 mock_insert_feature_pub: unittest.mock.Mock,
                                 mock_delete_feature_pub: unittest.mock.Mock):
        # Tests the function transferring data from a GFF record to the 'feature_pub' table
        self.assertIs(mock_query, self.client.query_all)
        self.assertIs(mock_pub, pub.Pub)
        self.assertIs(mock_insert_pub, self.client._handle_pub)
        self.assertIs(mock_featurepub, sequence.FeaturePub)
        self.assertIs(mock_insert_feature_pub, self.client._handle_feature_pub)
        self.assertIs(mock_delete_feature_pub, self.client._delete_feature_pub)

        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testname", feature_id=12)
        mock_insert_pub.return_value = utils.EmptyObject(pub_id=32, uniquename="")

        all_pubs = self.client._handle_publications(self.default_gff_record, feature_entry)
        mock_query.assert_called_with(sequence.FeaturePub, feature_id=12)
        mock_pub.assert_any_call(uniquename="PMID:12334", type_id=71)
        self.assertEqual(mock_insert_pub.call_count, 1)
        mock_featurepub.assert_any_call(feature_id=12, pub_id=32)
        self.assertEqual(mock_insert_feature_pub.call_count, 1)
        mock_delete_feature_pub.assert_called()
        self.assertEqual(len(all_pubs), 1)

    @unittest.mock.patch("pychado.io.gff.GFFImportClient._delete_feature_relationship")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient._handle_feature_relationship")
    @unittest.mock.patch("pychado.orm.sequence.FeatureRelationship")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient.query_first")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient.query_feature_relationship_by_type")
    def test_handle_relationships(self, mock_query: unittest.mock.Mock, mock_query_first: unittest.mock.Mock,
                                  mock_relationship: unittest.mock.Mock,
                                  mock_insert_relationship: unittest.mock.Mock,
                                  mock_delete_relationship: unittest.mock.Mock):
        # Tests the function transferring data from a GFF record to the 'feature_relationship' table
        self.assertIs(mock_query, self.client.query_feature_relationship_by_type)
        self.assertIs(mock_query_first, self.client.query_first)
        self.assertIs(mock_relationship, sequence.FeatureRelationship)
        self.assertIs(mock_insert_relationship, self.client._handle_feature_relationship)
        self.assertIs(mock_delete_relationship, self.client._delete_feature_relationship)

        subject_entry = sequence.Feature(organism_id=11, type_id=300, uniquename="testid", feature_id=33)
        object_entry = sequence.Feature(organism_id=11, type_id=400, uniquename="testparent", feature_id=44)
        mock_query_first.return_value = object_entry

        all_features = {object_entry.uniquename: object_entry}
        all_relationships = self.client._handle_relationships(self.default_gff_record, subject_entry, all_features)
        mock_query.assert_called_with(33, [62, 63])
        mock_query_first.assert_not_called()
        mock_relationship.assert_any_call(subject_id=33, object_id=44, type_id=62)
        self.assertEqual(mock_insert_relationship.call_count, 1)
        mock_delete_relationship.assert_called()
        self.assertEqual(len(all_relationships), 1)

        mock_relationship.reset_mock()
        mock_insert_relationship.reset_mock()
        mock_delete_relationship.reset_mock()
        all_features = {}
        all_relationships = self.client._handle_relationships(self.default_gff_record, subject_entry, all_features)
        mock_query.assert_called_with(33, [62, 63])
        mock_query_first.assert_called_with(sequence.Feature, organism_id=11, uniquename="testparent")
        mock_relationship.assert_any_call(subject_id=33, object_id=44, type_id=62)
        self.assertEqual(mock_insert_relationship.call_count, 1)
        mock_delete_relationship.assert_called()
        self.assertEqual(len(all_relationships), 1)

    @unittest.mock.patch("pychado.io.gff.GFFImportClient._delete_featureprop")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient._handle_featureprop")
    @unittest.mock.patch("pychado.orm.sequence.FeatureProp")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient.query_featureprop_by_type")
    def test_handle_properties(self, mock_query: unittest.mock.Mock, mock_prop: unittest.mock.Mock,
                               mock_insert_prop: unittest.mock.Mock, mock_delete_prop: unittest.mock.Mock):
        # Tests the function transferring data from a GFF record to the 'featureprop' table
        self.assertIs(mock_query, self.client.query_featureprop_by_type)
        self.assertIs(mock_prop, sequence.FeatureProp)
        self.assertIs(mock_insert_prop, self.client._handle_featureprop)
        self.assertIs(mock_delete_prop, self.client._delete_featureprop)

        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testname", feature_id=12)
        all_properties = self.client._handle_properties(self.default_gff_record, feature_entry)
        mock_query.assert_called_with(12, [51, 52, 53])
        mock_prop.assert_any_call(feature_id=12, type_id=51, value="3.5")
        mock_prop.assert_any_call(feature_id=12, type_id=52, value="testsource")
        mock_prop.assert_any_call(feature_id=12, type_id=53, value="testnote")
        self.assertEqual(mock_insert_prop.call_count, 3)
        mock_delete_prop.assert_called()
        self.assertEqual(len(all_properties), 3)

    @unittest.mock.patch("pychado.io.gff.GFFImportClient._delete_feature_dbxref")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient._handle_feature_dbxref")
    @unittest.mock.patch("pychado.orm.sequence.FeatureDbxRef")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient._handle_dbxref")
    @unittest.mock.patch("pychado.orm.general.DbxRef")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient._handle_db")
    @unittest.mock.patch("pychado.orm.general.Db")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient.query_all")
    def test_handle_crossrefs(self, mock_query: unittest.mock.Mock, mock_db: unittest.mock.Mock,
                              mock_insert_db: unittest.mock.Mock, mock_dbxref: unittest.mock.Mock,
                              mock_insert_dbxref: unittest.mock.Mock, mock_feature_dbxref: unittest.mock.Mock,
                              mock_insert_feature_dbxref: unittest.mock.Mock,
                              mock_delete_feature_dbxref: unittest.mock.Mock):
        # Tests the function transferring data from a GFF record to the 'feature_dbxref' table
        self.assertIs(mock_query, self.client.query_all)
        self.assertIs(mock_db, general.Db)
        self.assertIs(mock_insert_db, self.client._handle_db)
        self.assertIs(mock_dbxref, general.DbxRef)
        self.assertIs(mock_insert_dbxref, self.client._handle_dbxref)
        self.assertIs(mock_feature_dbxref, sequence.FeatureDbxRef)
        self.assertIs(mock_insert_feature_dbxref, self.client._handle_feature_dbxref)
        self.assertIs(mock_delete_feature_dbxref, self.client._delete_feature_dbxref)

        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testname", feature_id=12)
        mock_insert_db.return_value = utils.EmptyObject(db_id=44, name="")
        mock_insert_dbxref.return_value = utils.EmptyObject(dbxref_id=55, accession="", version="")

        all_crossrefs = self.client._handle_cross_references(self.default_gff_record, feature_entry)
        mock_query.assert_called_with(sequence.FeatureDbxRef, feature_id=12)
        mock_db.assert_any_call(name="testdb")
        mock_dbxref.assert_any_call(db_id=44, accession="testaccession", version="")
        mock_feature_dbxref.assert_any_call(feature_id=12, dbxref_id=55)
        self.assertEqual(mock_insert_db.call_count, 1)
        self.assertEqual(mock_insert_dbxref.call_count, 1)
        self.assertEqual(mock_insert_feature_dbxref.call_count, 1)
        mock_delete_feature_dbxref.assert_called()
        self.assertEqual(len(all_crossrefs), 1)

    @unittest.mock.patch("pychado.io.gff.GFFImportClient._delete_feature_cvterm")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient._handle_feature_cvterm")
    @unittest.mock.patch("pychado.orm.sequence.FeatureCvTerm")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient.query_first")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient.query_feature_cvterm_by_ontology")
    def test_handle_ontology_terms(self, mock_query: unittest.mock.Mock, mock_query_first: unittest.mock.Mock,
                                   mock_feature_cvterm: unittest.mock.Mock,
                                   mock_insert_feature_cvterm: unittest.mock.Mock,
                                   mock_delete_feature_cvterm: unittest.mock.Mock):
        # Tests the function transferring data from a GFF record to the 'feature_cvterm' table
        self.assertIs(mock_query, self.client.query_feature_cvterm_by_ontology)
        self.assertIs(mock_query_first, self.client.query_first)
        self.assertIs(mock_feature_cvterm, sequence.FeatureCvTerm)
        self.assertIs(mock_insert_feature_cvterm, self.client._handle_feature_cvterm)
        self.assertIs(mock_delete_feature_cvterm, self.client._delete_feature_cvterm)

        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testname", feature_id=12)
        mock_query_first.side_effect = [utils.EmptyObject(db_id=33),
                                        utils.EmptyObject(dbxref_id=44),
                                        utils.EmptyObject(cvterm_id=55, name="")]

        all_ontology_terms = self.client._handle_ontology_terms(self.default_gff_record, feature_entry)
        mock_query.assert_called_with(12, 131)
        mock_query_first.assert_any_call(general.Db, name="GO")
        mock_query_first.assert_any_call(general.DbxRef, db_id=33, accession="7890")
        mock_query_first.assert_any_call(cv.CvTerm, dbxref_id=44)
        self.assertEqual(mock_query_first.call_count, 3)
        mock_feature_cvterm.assert_any_call(feature_id=12, cvterm_id=55, pub_id=33)
        self.assertEqual(mock_insert_feature_cvterm.call_count, 1)
        mock_delete_feature_cvterm.assert_called()
        self.assertEqual(len(all_ontology_terms), 1)

    @unittest.mock.patch("pychado.io.gff.GFFImportClient._insert_gff_record_into_database")
    @unittest.mock.patch("gffutils.Feature")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient.query_first")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient.query_parent_features")
    def test_handle_protein(self, mock_query: unittest.mock.Mock, mock_query_first: unittest.mock.Mock,
                            mock_record: unittest.mock.Mock, mock_insert: unittest.mock.Mock):
        # Tests the function creating a new GFF record for a polypeptide and inserting it into the database
        self.assertIs(mock_query, self.client.query_parent_features)
        self.assertIs(mock_query_first, self.client.query_first)
        self.assertIs(mock_record, gffutils.Feature)
        self.assertIs(mock_insert, self.client._insert_gff_record_into_database)

        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testname", feature_id=12)
        organism_entry = organism.Organism(genus="", species="", abbreviation="testorganism", organism_id=1)
        all_features = {}

        # No attribute "protein_source_id"
        self.client._handle_protein(self.default_gff_record, feature_entry, organism_entry, all_features)
        mock_query.assert_not_called()
        mock_query_first.assert_not_called()
        mock_record.assert_not_called()
        mock_insert.assert_not_called()

        # Attribute "protein_source_id" present; GFF record is mRNA
        self.default_gff_record.featuretype = "mRNA"
        self.default_gff_record.attributes["protein_source_id"] = "testid"
        mock_query_first.return_value = sequence.FeatureLoc(feature_id=1, srcfeature_id=2, fmin=300, fmax=400, strand=1)
        self.client._handle_protein(self.default_gff_record, feature_entry, organism_entry, all_features)
        mock_query.assert_not_called()
        mock_query_first.assert_called_with(sequence.FeatureLoc, feature_id=12)
        mock_record.assert_called_with(seqid="testseqid", source="testsource", start=301, end=400, strand="+",
                                       featuretype="polypeptide", id="testid", attributes={"Derives_from": "testname"})
        mock_insert.assert_called()

        # Attribute "protein_source_id" present; GFF record is not mRNA
        self.default_gff_record.featuretype = "CDS"
        mock_query_object = mock_query.return_value
        mock_query_object.configure_mock(**{"first.return_value": sequence.Feature(
            organism_id=11, type_id=200, uniquename="othername", feature_id=13)})
        self.client._handle_protein(self.default_gff_record, feature_entry, organism_entry, all_features)
        mock_query.assert_called_with(12, [62])
        mock_query_first.assert_called_with(sequence.FeatureLoc, feature_id=13)
        mock_record.assert_called_with(seqid="testseqid", source="testsource", start=301, end=400, strand="+",
                                       featuretype="polypeptide", id="testid", attributes={"Derives_from": "othername"})
        mock_insert.assert_called()

    @unittest.mock.patch("pychado.io.gff.GFFImportClient._mark_feature_as_obsolete")
    @unittest.mock.patch("pychado.io.gff.GFFImportClient._load_feature_names")
    def test_mark_obsolete_features(self, mock_load: unittest.mock.Mock, mock_mark: unittest.mock.Mock):
        # Tests the function that marks features as obsolete if they are not present in a given dictionary
        self.assertIs(mock_load, self.client._load_feature_names)
        self.assertIs(mock_mark, self.client._mark_feature_as_obsolete)
        organism_entry = organism.Organism(genus="", species="", abbreviation="testorganism", organism_id=1)
        mock_load.return_value = ["id1", "id2", "id3", "seq"]
        self.client._mark_obsolete_features(
            organism_entry, {"id3": sequence.Feature(organism_id=1, type_id=1, uniquename="")}, ["seq"])
        mock_load.assert_called_with(organism_entry)
        mock_mark.assert_any_call(organism_entry, "id2")
        self.assertEqual(mock_mark.call_count, 2)

    def test_check_if_gff_attributes_are_recognized(self):
        # Tests the function that checks if all attributes of a GFF record are recognized
        feature = self.default_gff_record
        recognized = self.client._check_if_gff_attributes_are_recognized(feature)
        self.assertTrue(recognized)
        feature.attributes["other_attribute"] = "other_value"
        recognized = self.client._check_if_gff_attributes_are_recognized(feature)
        self.assertFalse(recognized)

    def test_create_feature(self):
        # Tests the function that creates an entry for the 'feature' table
        feature = self.client._create_feature(self.default_gff_record, 3, 5)
        self.assertEqual(feature.organism_id, 3)
        self.assertEqual(feature.type_id, 5)
        self.assertEqual(feature.uniquename, "testid")
        self.assertEqual(feature.name, "testname")
        self.assertEqual(feature.residues, "MCRA")
        self.assertEqual(feature.seqlen, 4)

    def test_create_featureloc(self):
        # Tests the function that creates an entry for the 'featureloc' table
        featureloc = self.client._create_featureloc(self.default_gff_record, 3, 5)
        self.assertEqual(featureloc.feature_id, 3)
        self.assertEqual(featureloc.srcfeature_id, 5)
        self.assertEqual(featureloc.fmin, 0)
        self.assertEqual(featureloc.fmax, 30)
        self.assertEqual(featureloc.strand, 1)
        self.assertEqual(featureloc.phase, 2)

    def test_extract_gff_name(self):
        # Tests the function that extracts the name from a GFF record
        feature = gffutils.Feature()
        name = self.client._extract_gff_name(feature)
        self.assertIsNone(name)
        feature.attributes = {"Name": "testname"}
        name = self.client._extract_gff_name(feature)
        self.assertEqual(name, "testname")
        feature.attributes = {"Name": ["othername"]}
        name = self.client._extract_gff_name(feature)
        self.assertEqual(name, "othername")

    def test_extract_gff_size(self):
        # Tests the function that extracts the sequence length from a GFF record
        feature = gffutils.Feature()
        size = self.client._extract_gff_size(feature)
        self.assertIsNone(size)
        feature.attributes = {"size": "587"}
        size = self.client._extract_gff_size(feature)
        self.assertEqual(size, 587)

    def test_extract_gff_synonyms(self):
        # Tests the function that extracts the synonyms from a GFF record
        feature = gffutils.Feature(attributes={"Alias": "testalias", "previous_systematic_id": ["testsynonym"]})
        synonyms = self.client._extract_gff_synonyms(feature)
        self.assertEqual(len(synonyms), 2)
        self.assertEqual(synonyms["alias"], ["testalias"])
        self.assertEqual(synonyms["previous_systematic_id"], ["testsynonym"])

    def test_extract_gff_residues(self):
        # Tests the function that extracts an amino acid sequence from a GFF record
        feature = gffutils.Feature()
        residues = self.client._extract_gff_translation(feature)
        self.assertIsNone(residues)
        feature.attributes = {"translation": "actgaa"}
        residues = self.client._extract_gff_translation(feature)
        self.assertEqual(residues, "ACTGAA")
        feature.attributes = {"translation": ["cTTg"]}
        residues = self.client._extract_gff_translation(feature)
        self.assertEqual(residues, "CTTG")

    def test_extract_gff_relationships(self):
        # Tests the function that extracts feature relationships from a GFF record
        feature = gffutils.Feature(attributes={"Parent": "parentterm", "Derives_from": ["otherterm"],
                                               "other_key": "other_value"})
        relationships = self.client._extract_gff_relationships(feature)
        self.assertEqual(len(relationships), 2)
        self.assertEqual(relationships["part_of"], ["parentterm"])
        self.assertEqual(relationships["derives_from"], ["otherterm"])

    def test_extract_gff_properties(self):
        # Tests the function that extracts feature properties from a GFF record
        feature = gffutils.Feature(source="testsource", score="2.54",
                                   attributes={"Parent": "parentterm", "comment": ["first_value", "second_value"]})
        properties = self.client._extract_gff_properties(feature)
        self.assertEqual(len(properties), 3)
        self.assertEqual(properties["source"], ["testsource"])
        self.assertEqual(properties["score"], ["2.54"])
        self.assertEqual(properties["comment"], ["first_value", "second_value"])

    def test_extract_gff_crossrefs(self):
        # Tests the function that extracts database cross references from a GFF record
        feature = gffutils.Feature(attributes={"Dbxref": "Wikipedia:gene", "Ontology_term": ["GO:12345", "GO:67890"]})
        crossrefs = self.client._extract_gff_crossrefs(feature)
        self.assertEqual(len(crossrefs), 1)
        self.assertIn("Wikipedia:gene", crossrefs)

    def test_extract_gff_ontology_terms(self):
        # Tests the function that extracts ontology terms from a GFF record
        feature = gffutils.Feature(attributes={"Dbxref": "Wikipedia:gene", "Ontology_term": ["GO:12345", "GO:67890"]})
        ontology_terms = self.client._extract_gff_ontology_terms(feature)
        self.assertEqual(len(ontology_terms), 2)
        self.assertIn("GO:67890", ontology_terms)

    def test_extract_gff_publications(self):
        # Tests the function that extracts publications from a GFF record
        feature = gffutils.Feature(attributes={"literature": "PMID:12345"})
        publications = self.client._extract_gff_publications(feature)
        self.assertEqual(len(publications), 1)
        self.assertIn("PMID:12345", publications)

    def test_extract_gff_protein_source_id(self):
        # Tests the function that extracts the ID of a polypeptide from a GFF record
        feature = gffutils.Feature()
        protein_id = self.client._extract_protein_source_id(feature)
        self.assertIsNone(protein_id)
        feature.attributes = {"protein_source_id": "testid"}
        protein_id = self.client._extract_protein_source_id(feature)
        self.assertEqual(protein_id, "testid")

    def test_extract_gff_sequence_names(self):
        # Tests the function that extracts sequence names from a GFF file
        gff_file = os.path.join(data_dir, 'gff_without_fasta.gff3')
        self.assertTrue(os.path.exists(gff_file))
        gff_db = self.client._create_sqlite_db(gff_file)
        sequence_names = self.client._extract_gff_sequence_names(gff_db)
        self.assertIn("CM000574", sequence_names)
        gff_db_name = self.client._sqlite_databases[0]
        os.remove(gff_db_name)
        self.assertFalse(os.path.exists(gff_db_name))


class TestGFFExport(unittest.TestCase):
    """Tests various functions used to export a GFF file from a database"""

    @classmethod
    def setUpClass(cls):
        # Creates an instance of the base class to be tested and instantiates global attributes
        cls.client = gff.GFFExportClient("testuri", test_environment=True)
        cls.client._parent_terms = {
            "part_of": cv.CvTerm(cv_id=6, dbxref_id=62, name="part_of", cvterm_id=62, is_relationshiptype=1),
            "derives_from": cv.CvTerm(cv_id=6, dbxref_id=63, name="derives_from", cvterm_id=63, is_relationshiptype=1)
        }
        cls.client._top_level_term = cv.CvTerm(cv_id=11, dbxref_id=91, name="top_level_seq", cvterm_id=91)
        cls.client._parent_type_ids = [62, 63]
        cls.client._go_db = general.Db(db_id=131, name="GO")

    @unittest.mock.patch("pychado.io.gff.GFFExportClient._export_gff_record")
    @unittest.mock.patch("pychado.io.gff.GFFExportClient.query_child_features")
    def test_handle_child_features(self, mock_query: unittest.mock.Mock, mock_export: unittest.mock.Mock):
        # Tests the function that exports GFF records for the child features of a given feature
        self.assertIs(mock_query, self.client.query_child_features)
        self.assertIs(mock_export, self.client._export_gff_record)

        feature_entry = sequence.Feature(feature_id=77, organism_id=11, type_id=200, uniquename="parentid")
        childfeature_entry = sequence.Feature(feature_id=78, organism_id=11, type_id=300, uniquename="childid")
        derivedfeature_entry = sequence.Feature(feature_id=79, organism_id=11, type_id=400, uniquename="derivedid")
        mock_query_obj = mock_query.return_value
        mock_query_obj.configure_mock(**{"all.side_effect": [[childfeature_entry],
                                                             [childfeature_entry, derivedfeature_entry]]})

        self.client._handle_child_features(feature_entry, "testsequence", None)
        self.assertIn(unittest.mock.call(77, 62), mock_query.mock_calls)
        self.assertIn(unittest.mock.call(77, 63), mock_query.mock_calls)
        self.assertEqual(mock_query.call_count, 2)
        self.assertIn(unittest.mock.call(childfeature_entry, "testsequence", {"part_of": "parentid"}, None),
                      mock_export.mock_calls)
        self.assertIn(unittest.mock.call(childfeature_entry, "testsequence", {"derives_from": "parentid"}, None),
                      mock_export.mock_calls)
        self.assertIn(unittest.mock.call(derivedfeature_entry, "testsequence", {"derives_from": "parentid"}, None),
                      mock_export.mock_calls)
        self.assertEqual(mock_export.call_count, 3)

    @unittest.mock.patch("pychado.io.fasta.FastaExportClient")
    @unittest.mock.patch("pychado.io.gff.GFFExportClient._append_fasta")
    def test_export_fasta(self, mock_append: unittest.mock.Mock, mock_fasta: unittest.mock.Mock):
        # Tests the export of FASTA sequences along with a GFF file
        self.assertIs(mock_append, self.client._append_fasta)
        self.assertIs(mock_fasta, fasta.FastaExportClient)

        # Output FASTA in separate file
        self.client._export_fasta("testgff", "testfasta", "testorganism")
        mock_fasta.assert_called_with("testuri", False)
        self.assertIn(unittest.mock.call().export("testfasta", "testorganism", "contigs", ""), mock_fasta.mock_calls)
        mock_append.assert_not_called()

        # Append FASTA to GFF file
        mock_fasta.reset_mock()
        mock_append.reset_mock()
        self.client._export_fasta("testgff", "", "testorganism")
        mock_fasta.assert_called_with("testuri", False)
        mock_append.assert_called()

    def test_append_fasta(self):
        # Tests the function joining a GFF file with a FASTA file
        gff_file = os.path.join(data_dir, 'gff_without_fasta.gff3')
        fasta_file = os.path.join(data_dir, 'fasta_only.fa')
        joined_file = tempfile.mkstemp()[1]
        shutil.copy(gff_file, joined_file)
        actual_joined_file = os.path.join(data_dir, 'gff_with_fasta.gff3')
        self.client._append_fasta(joined_file, fasta_file)
        self.assertTrue(filecmp.cmp(joined_file, actual_joined_file))
        os.remove(joined_file)

    def test_write_gff_header(self):
        # Tests the correct creation of GFF file headers
        header_file = tempfile.mkstemp()[1]
        header_file_handle = utils.open_file_write(header_file)
        self.client._write_gff_header(header_file_handle, [sequence.Feature(feature_id=77, organism_id=11, type_id=200,
                                                                            seqlen=11697295, uniquename="CM000574")])
        utils.close(header_file_handle)
        actual_header_file = os.path.join(data_dir, 'gff_header.gff3')
        self.assertTrue(filecmp.cmp(header_file, actual_header_file))
        os.remove(header_file)

    def test_print_gff_record(self):
        # Tests the correct printing of GFF records to file
        gff_record = gffutils.Feature(seqid="CM000574", id="FGSG_11579", source="chado", featuretype="gene",
                                      start=1517, end=2509, strand="+", attributes={"ID": ["FGSG_11579"]})
        file = tempfile.mkstemp()[1]
        file_handle = utils.open_file_write(file)
        self.client._print_gff_record(gff_record, file_handle)
        utils.close(file_handle)
        actual_file = os.path.join(data_dir, 'gff_line.gff3')
        self.assertTrue(filecmp.cmp(file, actual_file))
        os.remove(file)

    @unittest.mock.patch("pychado.io.gff.GFFExportClient.query_first")
    def test_create_gff_record(self, mock_query: unittest.mock.Mock):
        # Tests the function that creates a GFF record
        self.assertIs(mock_query, self.client.query_first)
        feature_entry = sequence.Feature(feature_id=77, organism_id=11, type_id=200, seqlen=66, uniquename="testid",
                                         name="testname")

        # Create GFF record for top-level feature
        mock_query.return_value = None
        gff_record = self.client._create_gff_record(feature_entry, "testsequence")
        self.assertEqual(gff_record.seqid, "testsequence")
        self.assertEqual(gff_record.id, "testid")
        self.assertEqual(gff_record.start, 1)
        self.assertEqual(gff_record.end, 66)
        self.assertEqual(gff_record.strand, ".")
        self.assertEqual(gff_record.frame, ".")
        self.assertEqual(gff_record.attributes["ID"], ["testid"])
        self.assertEqual(gff_record.attributes["Name"], ["testname"])

        # Create GFF record for feature located on a sequence
        mock_query.return_value = sequence.FeatureLoc(feature_id=1, srcfeature_id=2, fmin=10, fmax=100, strand=1,
                                                      phase=2)
        gff_record = self.client._create_gff_record(feature_entry, "testsequence")
        self.assertEqual(gff_record.seqid, "testsequence")
        self.assertEqual(gff_record.id, "testid")
        self.assertEqual(gff_record.start, 11)
        self.assertEqual(gff_record.end, 100)
        self.assertEqual(gff_record.strand, "+")
        self.assertEqual(gff_record.frame, "2")

    @unittest.mock.patch("pychado.io.gff.GFFExportClient.query_parent_features")
    def test_has_feature_parents(self, mock_query: unittest.mock.Mock):
        # Tests the function that checks if a feature entry has parents
        self.assertIs(mock_query, self.client.query_parent_features)
        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testid", feature_id=77)
        mock_query_obj = mock_query.return_value
        mock_query_obj.configure_mock(**{"first.return_value": None})
        has_parents = self.client._has_feature_parents(feature_entry)
        mock_query.assert_called_with(77, [62, 63])
        self.assertFalse(has_parents)
        mock_query_obj.configure_mock(**{"first.return_value": sequence.Feature(
            organism_id=11, type_id=300, uniquename="parentid", feature_id=88)})
        has_parents = self.client._has_feature_parents(feature_entry)
        self.assertTrue(has_parents)

    @unittest.mock.patch("pychado.io.gff.GFFExportClient.query_first")
    def test_extract_feature_type(self, mock_query: unittest.mock.Mock):
        # Tests the function that extracts the type of a feature entry from the relevant database table
        self.assertIs(mock_query, self.client.query_first)
        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testid", feature_id=77)
        mock_query.return_value = cv.CvTerm(cv_id=1, dbxref_id=2, name="testtype")
        featuretype = self.client._extract_feature_type(feature_entry)
        mock_query.assert_called_with(cv.CvTerm, cvterm_id=200)
        self.assertEqual(featuretype, "testtype")

    @unittest.mock.patch("pychado.io.gff.GFFExportClient.query_feature_synonyms")
    def test_extract_feature_synonyms(self, mock_query: unittest.mock.Mock):
        # Tests the function that extracts the synonyms of a feature entry from the relevant database table
        self.assertIs(mock_query, self.client.query_feature_synonyms)
        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testid", feature_id=77)
        mock_query_obj = mock_query.return_value
        mock_query_obj.configure_mock(**{"all.return_value": [("sometype", "somename"), ("sometype", "othername")]})
        synonyms = self.client._extract_feature_synonyms(feature_entry)
        mock_query.assert_called_with(77)
        self.assertEqual(synonyms, {"sometype": ["somename", "othername"]})

    @unittest.mock.patch("pychado.io.gff.GFFExportClient.query_feature_properties")
    def test_extract_feature_properties(self, mock_query: unittest.mock.Mock):
        # Tests the function that extracts the synonyms of a feature entry from the relevant database table
        self.assertIs(mock_query, self.client.query_feature_properties)
        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testid", feature_id=77)
        mock_query_obj = mock_query.return_value
        mock_query_obj.configure_mock(**{"all.return_value": [("somekey", "somevalue"), ("somekey", "othervalue"),
                                                              ("otherkey", "yetanothervalue")]})
        properties = self.client._extract_feature_properties(feature_entry)
        mock_query.assert_called_with(77)
        self.assertEqual(properties, {"somekey": ["somevalue", "othervalue"], "otherkey": ["yetanothervalue"]})

    @unittest.mock.patch("pychado.io.gff.GFFExportClient.query_feature_pubs")
    def test_extract_feature_publications(self, mock_query: unittest.mock.Mock):
        # Tests the function that extracts the publications of a feature entry from the relevant database table
        self.assertIs(mock_query, self.client.query_feature_pubs)
        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testid", feature_id=77)
        mock_query_obj = mock_query.return_value
        mock_query_obj.configure_mock(**{"all.return_value": [("somepublication", ), ("otherpublication", )]})
        publications = self.client._extract_feature_publications(feature_entry)
        mock_query.assert_called_with(77)
        self.assertEqual(publications, ["somepublication", "otherpublication"])

    @unittest.mock.patch("pychado.io.gff.GFFExportClient.query_feature_dbxrefs")
    def test_extract_feature_dbxrefs(self, mock_query: unittest.mock.Mock):
        # Tests the function that extracts the cross references of a feature entry from the relevant database table
        self.assertIs(mock_query, self.client.query_feature_dbxrefs)
        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testid", feature_id=77)
        mock_query_obj = mock_query.return_value
        mock_query_obj.configure_mock(**{"all.return_value": [("somedb", "someacc",), ("otherdb", "otheracc")]})
        dbxrefs = self.client._extract_feature_cross_references(feature_entry)
        mock_query.assert_called_with(77)
        self.assertEqual(dbxrefs, ["somedb:someacc", "otherdb:otheracc"])

    @unittest.mock.patch("pychado.io.gff.GFFExportClient.query_feature_ontology_terms")
    def test_extract_feature_ontology_terms(self, mock_query: unittest.mock.Mock):
        # Tests the function that extracts the ontology terms of a feature entry from the relevant database table
        self.assertIs(mock_query, self.client.query_feature_ontology_terms)
        feature_entry = sequence.Feature(organism_id=11, type_id=200, uniquename="testid", feature_id=77)
        mock_query_obj = mock_query.return_value
        mock_query_obj.configure_mock(**{"all.return_value": [("GO", "12345"), ("SO", "54321")]})
        ontology_terms = self.client._extract_feature_ontology_terms(feature_entry)
        mock_query.assert_called_with(77, 131)
        self.assertEqual(ontology_terms, ["GO:12345", "SO:54321"])

    def test_add_gff_featuretype(self):
        # Tests the function that adds the 'type' and the attribute 'translation' to a GFF record
        gff_record = gffutils.Feature()
        self.client._add_gff_featuretype(gff_record, "gene", "MKGHU")
        self.assertEqual(gff_record.featuretype, "gene")
        self.assertNotIn("translation", gff_record.attributes)
        self.client._add_gff_featuretype(gff_record, "polypeptide", "MKGHU")
        self.assertEqual(gff_record.featuretype, "polypeptide")
        self.assertEqual(gff_record.attributes["translation"], ["MKGHU"])

    def test_add_gff_synonyms(self):
        # Tests the function that adds the attribute 'Alias' and various related attributes to a GFF record
        gff_record = gffutils.Feature()
        self.client._add_gff_synonyms(gff_record, {"alias": ["testalias", "otheralias"], "synonym": ["testsynonym"],
                                                   "otherkey": ["othervalue"]})
        self.assertEqual(gff_record.attributes["Alias"], ["testalias", "otheralias"])
        self.assertEqual(gff_record.attributes["synonym"], ["testsynonym"])
        self.assertNotIn("otherkey", gff_record.attributes)

    def test_add_gff_properties(self):
        # Tests the function that adds the 'source', 'score' and various attributes to a GFF record
        gff_record = gffutils.Feature()
        self.client._add_gff_properties(
            gff_record, {"source": ["testsource"], "score": ["testscore"], "comment": ["testcomment", "othercomment"],
                         "otherkey": ["othervalue"]})
        self.assertEqual(gff_record.source, "testsource")
        self.assertEqual(gff_record.score, "testscore")
        self.assertEqual(gff_record.attributes["comment"], ["testcomment", "othercomment"])
        self.assertNotIn("otherkey", gff_record.attributes)

    def test_add_gff_publications(self):
        # Tests the function that adds the attribute 'literature' to a GFF record
        gff_record = gffutils.Feature()
        self.client._add_gff_publications(gff_record, ["testpub"])
        self.assertEqual(gff_record.attributes["literature"], ["testpub"])

    def test_add_gff_cross_references(self):
        # Tests the function that adds the attribute 'Dbxref' to a GFF record
        gff_record = gffutils.Feature()
        self.client._add_gff_cross_references(gff_record, ["testdbxref", "otherdbxref"])
        self.assertEqual(gff_record.attributes["Dbxref"], ["testdbxref", "otherdbxref"])

    def test_add_gff_ontology_terms(self):
        # Tests the function that adds the attribute 'Ontology_term' to a GFF record
        gff_record = gffutils.Feature()
        self.client._add_gff_ontology_terms(gff_record, ["GO:12345"])
        self.assertEqual(gff_record.attributes["Ontology_term"], ["GO:12345"])

    def test_add_gff_relationships(self):
        # Tests the function that adds the attributes 'Parent' and 'Derives_from' to a GFF record
        gff_record = gffutils.Feature()
        self.client._add_gff_relationships(gff_record, {"part_of": "feature1", "derives_from": "feature2",
                                                        "orthologous_to": "other_feature"})
        self.assertEqual(gff_record.attributes["Parent"], ["feature1"])
        self.assertEqual(gff_record.attributes["Derives_from"], ["feature2"])
        self.assertNotIn("orthologous_to", gff_record.attributes)


class TestGFFFunctions(unittest.TestCase):
    """Tests various general functions used in connection with GFF import/export"""

    @classmethod
    def setUpClass(cls):
        # Creates an instance of the base class to be tested
        cls.client = gff.GFFClient()

    def test_convert_strand(self):
        # Tests the function converting the 'strand' attribute from string notation to integer notation
        gff_strand = "+"
        chado_strand = self.client.convert_strand(gff_strand)
        self.assertEqual(chado_strand, 1)
        gff_strand = "-"
        chado_strand = self.client.convert_strand(gff_strand)
        self.assertEqual(chado_strand, -1)
        gff_strand = "something_else"
        chado_strand = self.client.convert_strand(gff_strand)
        self.assertIsNone(chado_strand)

    def test_back_convert_strand(self):
        # Tests the function converting the 'strand' attribute from integer notation to string notation
        chado_strand = 1
        gff_strand = self.client.back_convert_strand(chado_strand)
        self.assertEqual(gff_strand, "+")
        chado_strand = -5
        gff_strand = self.client.back_convert_strand(chado_strand)
        self.assertEqual(gff_strand, "-")
        chado_strand = None
        gff_strand = self.client.back_convert_strand(chado_strand)
        self.assertEqual(gff_strand, ".")

    def test_convert_frame(self):
        # Tests the function converting the 'frame' attribute from string notation to integer notation
        gff_frame = "."
        chado_frame = self.client.convert_frame(gff_frame)
        self.assertIsNone(chado_frame)
        gff_frame = "2"
        chado_frame = self.client.convert_frame(gff_frame)
        self.assertEqual(chado_frame, 2)
        gff_frame = "3"
        chado_frame = self.client.convert_frame(gff_frame)
        self.assertIsNone(chado_frame)

    def test_back_convert_frame(self):
        # Tests the function converting the 'frame' attribute from integer notation to string notation
        chado_frame = 1
        gff_frame = self.client.back_convert_frame(chado_frame)
        self.assertEqual(gff_frame, "1")
        chado_frame = None
        gff_frame = self.client.back_convert_frame(chado_frame)
        self.assertEqual(gff_frame, ".")


if __name__ == '__main__':
    unittest.main(verbosity=2, buffer=True)



import sqlalchemy.orm
import sqlalchemy.sql.functions
from . import base, general, cv, organism, pub

# Object-relational mappings for the CHADO Sequence/Feature module


class Feature(base.PublicBase):
    """Class for the CHADO 'feature' table"""
    # Columns
    feature_id = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=False, primary_key=True, autoincrement=True)
    dbxref_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        general.DbxRef.dbxref_id, onupdate="CASCADE", ondelete="SET NULL"), nullable=True)
    organism_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        organism.Organism.organism_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    name = sqlalchemy.Column(sqlalchemy.VARCHAR(255), nullable=True)
    uniquename = sqlalchemy.Column(sqlalchemy.TEXT, nullable=False)
    residues = sqlalchemy.Column(sqlalchemy.TEXT, nullable=True)
    seqlen = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=True)
    md5checksum = sqlalchemy.Column(sqlalchemy.CHAR(32), nullable=True)
    type_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        cv.CvTerm.cvterm_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    is_analysis = sqlalchemy.Column(sqlalchemy.BOOLEAN, nullable=False, server_default="False")
    is_obsolete = sqlalchemy.Column(sqlalchemy.BOOLEAN, nullable=False, server_default="False")
    timeaccessioned = sqlalchemy.Column(sqlalchemy.TIMESTAMP, nullable=False,
                                        server_default=sqlalchemy.sql.functions.now())
    timelastmodified = sqlalchemy.Column(sqlalchemy.TIMESTAMP, nullable=False,
                                         server_default=sqlalchemy.sql.functions.now())

    # Constraints
    __tablename__ = "feature"
    __table_args__ = (sqlalchemy.UniqueConstraint(organism_id, uniquename, type_id, name="feature_c1"),
                      sqlalchemy.Index("feature_idx1", dbxref_id),
                      sqlalchemy.Index("feature_idx2", organism_id),
                      sqlalchemy.Index("feature_idx3", type_id),
                      sqlalchemy.Index("feature_idx4", uniquename),
                      sqlalchemy.Index("feature_idx5", sqlalchemy.func.lower(name)))

    # Relationships
    dbxref = sqlalchemy.orm.relationship(general.DbxRef, foreign_keys=dbxref_id, backref="feature_dbxref")
    organism = sqlalchemy.orm.relationship(organism.Organism, foreign_keys=organism_id, backref="feature_organism")
    type = sqlalchemy.orm.relationship(cv.CvTerm, foreign_keys=type_id, backref="feature_type")

    # Initialisation
    def __init__(self, organism_id, type_id, uniquename, dbxref_id=None, name=None, residues=None, seqlen=None,
                 md5checksum=None, is_analysis=False, is_obsolete=False, timeaccessioned=sqlalchemy.sql.functions.now(),
                 timelastmodified=sqlalchemy.sql.functions.now(), feature_id=None):
        for key, value in locals().items():
            if key != self:
                setattr(self, key, value)

    # Representation
    def __repr__(self):
        return "<sequence.Feature(feature_id={0}, dbxref_id={1}, organism_id={2}, name='{3}', uniquename='{4}', " \
               "seqlen={5}, type_id={6}, is_analysis={7}, is_obsolete={8}, timeaccessioned='{9}', " \
               "timelastmodified='{10}')>"\
            .format(self.feature_id, self.dbxref_id, self.organism_id, self.name, self.uniquename, self.seqlen,
                    self.type_id, self.is_analysis, self.is_obsolete, self.timeaccessioned, self.timelastmodified)


class FeatureCvTerm(base.PublicBase):
    """Class for the CHADO 'feature_cvterm' table"""
    # Columns
    feature_cvterm_id = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=False, primary_key=True, autoincrement=True)
    feature_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        Feature.feature_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    cvterm_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        cv.CvTerm.cvterm_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    pub_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        pub.Pub.pub_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    is_not = sqlalchemy.Column(sqlalchemy.BOOLEAN, nullable=False, server_default="False")
    rank = sqlalchemy.Column(sqlalchemy.INTEGER, nullable=False, server_default="0")

    # Constraints
    __tablename__ = "feature_cvterm"
    __table_args__ = (sqlalchemy.UniqueConstraint(feature_id, cvterm_id, pub_id, rank, name="feature_cvterm_c1"),
                      sqlalchemy.Index("feature_cvterm_idx1", feature_id),
                      sqlalchemy.Index("feature_cvterm_idx2", cvterm_id),
                      sqlalchemy.Index("feature_cvterm_idx3", pub_id))

    # Relationships
    feature = sqlalchemy.orm.relationship(Feature, foreign_keys=feature_id, backref="feature_cvterm_feature")
    cvterm = sqlalchemy.orm.relationship(cv.CvTerm, foreign_keys=cvterm_id, backref="feature_cvterm_cvterm")
    pub = sqlalchemy.orm.relationship(pub.Pub, foreign_keys=pub_id, backref="feature_cvterm_pub")

    # Initialisation
    def __init__(self, feature_id, cvterm_id, pub_id, is_not=None, rank=0, feature_cvterm_id=None):
        for key, value in locals().items():
            if key != self:
                setattr(self, key, value)

    # Representation
    def __repr__(self):
        return "<sequence.FeatureCvTerm(feature_cvterm_id={0}, feature_id={1}, cvterm_id={2}, pub_id={3}, is_not={4}, "\
               "rank={5})>".format(self.feature_cvterm_id, self.feature_id, self.cvterm_id, self.pub_id, self.is_not,
                                   self.rank)


class FeatureCvTermDbxRef(base.PublicBase):
    """Class for the CHADO 'feature_cvterm_dbxref' table"""
    # Columns
    feature_cvterm_dbxref_id = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=False, primary_key=True,
                                                 autoincrement=True)
    feature_cvterm_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        FeatureCvTerm.feature_cvterm_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    dbxref_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        general.DbxRef.dbxref_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)

    # Constraints
    __tablename__ = "feature_cvterm_dbxref"
    __table_args__ = (sqlalchemy.UniqueConstraint(feature_cvterm_id, dbxref_id, name="feature_cvterm_dbxref_c1"),
                      sqlalchemy.Index("feature_cvterm_dbxref_idx1", feature_cvterm_id),
                      sqlalchemy.Index("feature_cvterm_dbxref_idx2", dbxref_id))

    # Relationships
    feature_cvterm = sqlalchemy.orm.relationship(FeatureCvTerm, foreign_keys=feature_cvterm_id,
                                                 backref="feature_cvterm_dbxref_feature_cvterm")
    dbxref = sqlalchemy.orm.relationship(general.DbxRef, foreign_keys=dbxref_id, backref="feature_cvterm_dbxref_dbxref")

    # Initialisation
    def __init__(self, feature_cvterm_id, dbxref_id, feature_cvterm_dbxref_id=None):
        for key, value in locals().items():
            if key != self:
                setattr(self, key, value)

    # Representation
    def __repr__(self):
        return "<sequence.FeatureCvTermDbxRef(feature_cvterm_dbxref_id={0}, feature_cvterm_id={1}, dbxref_id={2})>".\
            format(self.feature_cvterm_dbxref_id, self.feature_cvterm_id, self.dbxref_id)


class FeatureCvTermProp(base.PublicBase):
    """Class for the CHADO 'feature_cvtermprop' table"""
    # Columns
    feature_cvtermprop_id = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=False, primary_key=True, autoincrement=True)
    feature_cvterm_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        FeatureCvTerm.feature_cvterm_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    type_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        cv.CvTerm.cvterm_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    value = sqlalchemy.Column(sqlalchemy.TEXT, nullable=True)
    rank = sqlalchemy.Column(sqlalchemy.INTEGER, nullable=False, server_default="0")

    # Constraints
    __tablename__ = "feature_cvtermprop"
    __table_args__ = (sqlalchemy.UniqueConstraint(feature_cvterm_id, type_id, rank, name="feature_cvtermprop_c1"),
                      sqlalchemy.Index("feature_cvtermprop_idx1", feature_cvterm_id),
                      sqlalchemy.Index("feature_cvtermprop_idx2", type_id))

    # Relationships
    feature_cvterm = sqlalchemy.orm.relationship(FeatureCvTerm, foreign_keys=feature_cvterm_id,
                                                 backref="feature_cvtermprop_organism")
    type = sqlalchemy.orm.relationship(cv.CvTerm, foreign_keys=type_id, backref="feature_cvtermprop_type")

    # Initialisation
    def __init__(self, feature_cvterm_id, type_id, value=None, rank=0, feature_cvtermprop_id=None):
        for key, val in locals().items():
            if key != self:
                setattr(self, key, val)

    # Representation
    def __repr__(self):
        return "<sequence.FeatureCvTermProp(feature_cvtermprop_id={0}, feature_cvterm_id={1}, type_id={2}, " \
               "value='{3}', rank={4})>".format(self.feature_cvtermprop_id, self.feature_cvterm_id, self.type_id,
                                                self.value, self.rank)


class FeatureCvTermPub(base.PublicBase):
    """Class for the CHADO 'feature_cvterm_pub' table"""
    # Columns
    feature_cvterm_pub_id = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=False, primary_key=True, autoincrement=True)
    feature_cvterm_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        FeatureCvTerm.feature_cvterm_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    pub_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        pub.Pub.pub_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)

    # Constraints
    __tablename__ = "feature_cvterm_pub"
    __table_args__ = (sqlalchemy.UniqueConstraint(feature_cvterm_id, pub_id, name="feature_cvterm_pub_c1"),
                      sqlalchemy.Index("feature_cvterm_pub_idx1", feature_cvterm_id),
                      sqlalchemy.Index("feature_cvterm_pub_idx2", pub_id))

    # Relationships
    feature_cvterm = sqlalchemy.orm.relationship(FeatureCvTerm, foreign_keys=feature_cvterm_id,
                                                 backref="feature_cvterm_pub_feature_cvterm")
    pub = sqlalchemy.orm.relationship(pub.Pub, foreign_keys=pub_id, backref="feature_cvterm_pub_pub")

    # Initialisation
    def __init__(self, feature_cvterm_id, pub_id, feature_cvterm_pub_id=None):
        for key, value in locals().items():
            if key != self:
                setattr(self, key, value)

    # Representation
    def __repr__(self):
        return "<sequence.FeatureCvTermPub(feature_cvterm_pub_id={0}, feature_cvterm_id={1}, pub_id={2})>".\
            format(self.feature_cvterm_pub_id, self.feature_cvterm_id, self.pub_id)


class FeatureDbxRef(base.PublicBase):
    """Class for the CHADO 'feature_dbxref' table"""
    # Columns
    feature_dbxref_id = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=False, primary_key=True, autoincrement=True)
    feature_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        Feature.feature_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    dbxref_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        general.DbxRef.dbxref_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    is_current = sqlalchemy.Column(sqlalchemy.BOOLEAN, nullable=False, server_default="True")

    # Constraints
    __tablename__ = "feature_dbxref"
    __table_args__ = (sqlalchemy.UniqueConstraint(feature_id, dbxref_id, name="feature_dbxref_c1"),
                      sqlalchemy.Index("feature_dbxref_idx1", feature_id),
                      sqlalchemy.Index("feature_dbxref_idx2", dbxref_id))

    # Relationships
    feature = sqlalchemy.orm.relationship(Feature, foreign_keys=feature_id, backref="feature_dbxref_feature")
    dbxref = sqlalchemy.orm.relationship(general.DbxRef, foreign_keys=dbxref_id, backref="feature_dbxref_dbxref")

    # Initialisation
    def __init__(self, feature_id, dbxref_id, is_current=True, feature_dbxref_id=None):
        for key, value in locals().items():
            if key != self:
                setattr(self, key, value)

    # Representation
    def __repr__(self):
        return "<sequence.FeatureDbxRef(feature_dbxref_id={0}, feature_id={1}, dbxref_id={2}, is_current={3})>".\
            format(self.feature_dbxref_id, self.feature_id, self.dbxref_id, self.is_current)


class FeaturePub(base.PublicBase):
    """Class for the CHADO 'feature_pub' table"""
    # Columns
    feature_pub_id = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=False, primary_key=True, autoincrement=True)
    feature_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        Feature.feature_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    pub_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        pub.Pub.pub_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)

    # Constraints
    __tablename__ = "feature_pub"
    __table_args__ = (sqlalchemy.UniqueConstraint(feature_id, pub_id, name="feature_pub_c1"),
                      sqlalchemy.Index("feature_pub_idx1", feature_id),
                      sqlalchemy.Index("feature_pub_idx2", pub_id))

    # Relationships
    feature = sqlalchemy.orm.relationship(Feature, foreign_keys=feature_id, backref="feature_pub_feature")
    pub = sqlalchemy.orm.relationship(pub.Pub, foreign_keys=pub_id, backref="feature_pub_pub")

    # Initialisation
    def __init__(self, feature_id, pub_id, feature_pub_id=None):
        for key, value in locals().items():
            if key != self:
                setattr(self, key, value)

    # Representation
    def __repr__(self):
        return "<sequence.FeaturePub(feature_pub_id={0}, feature_id={1}, pub_id={2})>". \
            format(self.feature_pub_id, self.feature_id, self.pub_id)


class FeaturePubProp(base.PublicBase):
    """Class for the CHADO 'feature_pubprop' table"""
    # Columns
    feature_pubprop_id = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=False, primary_key=True, autoincrement=True)
    feature_pub_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        FeaturePub.feature_pub_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    type_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        cv.CvTerm.cvterm_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    value = sqlalchemy.Column(sqlalchemy.TEXT, nullable=True)
    rank = sqlalchemy.Column(sqlalchemy.INTEGER, nullable=False, server_default="0")

    # Constraints
    __tablename__ = "feature_pubprop"
    __table_args__ = (sqlalchemy.UniqueConstraint(feature_pub_id, type_id, rank, name="feature_pubprop_c1"),
                      sqlalchemy.Index("feature_pubprop_idx1", feature_pub_id),
                      sqlalchemy.Index("feature_pubprop_idx2", type_id))

    # Relationships
    feature_pub = sqlalchemy.orm.relationship(FeaturePub, foreign_keys=feature_pub_id,
                                              backref="feature_pubprop_feature_pub")
    type = sqlalchemy.orm.relationship(cv.CvTerm, foreign_keys=type_id, backref="feature_pubprop_type")

    # Initialisation
    def __init__(self, feature_pub_id, type_id, value=None, rank=0, feature_pubprop_id=None):
        for key, val in locals().items():
            if key != self:
                setattr(self, key, val)

    # Representation
    def __repr__(self):
        return "<sequence.FeaturePubProp(feature_pubprop_id={0}, feature_pub_id={1}, type_id={2}, value='{3}', " \
               "rank={4})>".format(self.feature_pubprop_id, self.feature_pub_id, self.type_id, self.value, self.rank)


class FeatureRelationship(base.PublicBase):
    """Class for the CHADO 'feature_relationship' table"""
    # Columns
    feature_relationship_id = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=False, primary_key=True, autoincrement=True)
    subject_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        Feature.feature_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    object_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        Feature.feature_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    type_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        cv.CvTerm.cvterm_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    value = sqlalchemy.Column(sqlalchemy.TEXT, nullable=True)
    rank = sqlalchemy.Column(sqlalchemy.INTEGER, nullable=False, server_default="0")

    # Constraints
    __tablename__ = "feature_relationship"
    __table_args__ = (sqlalchemy.UniqueConstraint(subject_id, object_id, type_id, rank, name="feature_relationship_c1"),
                      sqlalchemy.Index("feature_relationship_idx1", subject_id),
                      sqlalchemy.Index("feature_relationship_idx2", object_id),
                      sqlalchemy.Index("feature_relationship_idx3", type_id))

    # Relationships
    type = sqlalchemy.orm.relationship(cv.CvTerm, foreign_keys=type_id, backref="feature_relationship_type")
    subject = sqlalchemy.orm.relationship(Feature, foreign_keys=subject_id, backref="feature_relationship_subject")
    object = sqlalchemy.orm.relationship(Feature, foreign_keys=object_id, backref="feature_relationship_object")

    # Initialisation
    def __init__(self, subject_id, object_id, type_id, value=None, rank=0, feature_relationship_id=None):
        for key, value in locals().items():
            if key != self:
                setattr(self, key, value)

    # Representation
    def __repr__(self):
        return "<sequence.FeatureRelationship(feature_relationship_id={0}, subject_id={1}, object_id={2}, " \
               "type_id={3}, value='{4}', rank={5})>".format(self.feature_relationship_id, self.subject_id,
                                                             self.object_id, self.type_id, self.value, self.rank)


class FeatureRelationshipPub(base.PublicBase):
    """Class for the CHADO 'feature_relationship_pub' table"""
    # Columns
    feature_relationship_pub_id = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=False, primary_key=True,
                                                    autoincrement=True)
    feature_relationship_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        FeatureRelationship.feature_relationship_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    pub_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        pub.Pub.pub_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)

    # Constraints
    __tablename__ = "feature_relationship_pub"
    __table_args__ = (sqlalchemy.UniqueConstraint(feature_relationship_id, pub_id, name="feature_relationship_pub_c1"),
                      sqlalchemy.Index("feature_relationship_pub_idx1", feature_relationship_id),
                      sqlalchemy.Index("feature_relationship_pub_idx2", pub_id))

    # Relationships
    feature_relationship = sqlalchemy.orm.relationship(FeatureRelationship, foreign_keys=feature_relationship_id,
                                                       backref="feature_relationship_pub_feature_relationship")
    pub = sqlalchemy.orm.relationship(pub.Pub, foreign_keys=pub_id, backref="feature_relationship_pub_pub")

    # Initialisation
    def __init__(self, feature_relationship_id, pub_id, feature_relationship_pub_id=None):
        for key, value in locals().items():
            if key != self:
                setattr(self, key, value)

    # Representation
    def __repr__(self):
        return "<sequence.FeatureRelationshipPub(feature_relationship_pub_id={0}, feature_relationship_id={1}, " \
               "pub_id={2})>".format(self.feature_relationship_pub_id, self.feature_relationship_id, self.pub_id)


class FeatureRelationshipProp(base.PublicBase):
    """Class for the CHADO 'feature_relationshipprop' table"""
    # Columns
    feature_relationshipprop_id = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=False, primary_key=True,
                                                    autoincrement=True)
    feature_relationship_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        FeatureRelationship.feature_relationship_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    type_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        cv.CvTerm.cvterm_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    value = sqlalchemy.Column(sqlalchemy.TEXT, nullable=True)
    rank = sqlalchemy.Column(sqlalchemy.INTEGER, nullable=False, server_default="0")

    # Constraints
    __tablename__ = "feature_relationshipprop"
    __table_args__ = (sqlalchemy.UniqueConstraint(feature_relationship_id, type_id, rank,
                                                  name="feature_relationshipprop_c1"),
                      sqlalchemy.Index("feature_relationshipprop_idx1", feature_relationship_id),
                      sqlalchemy.Index("feature_relationshipprop_idx2", type_id))

    # Relationships
    feature_relationship = sqlalchemy.orm.relationship(FeatureRelationship, foreign_keys=feature_relationship_id,
                                                       backref="feature_relationshipprop_feature_relationship")
    type = sqlalchemy.orm.relationship(cv.CvTerm, foreign_keys=type_id, backref="feature_relationshipprop_type")

    # Initialisation
    def __init__(self, feature_relationship_id, type_id, value=None, rank=0, feature_relationshipprop_id=None):
        for key, val in locals().items():
            if key != self:
                setattr(self, key, val)

    # Representation
    def __repr__(self):
        return "<sequence.FeatureRelationshipProp(feature_relationshipprop_id={0}, feature_relationship_id={1}, " \
               "type_id={2}, value='{3}', rank={4})>".\
            format(self.feature_relationshipprop_id, self.feature_relationship_id, self.type_id, self.value, self.rank)


class FeatureRelationshipPropPub(base.PublicBase):
    """Class for the CHADO 'feature_relationshipprop_pub' table"""
    # Columns
    feature_relationshipprop_pub_id = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=False, primary_key=True,
                                                        autoincrement=True)
    feature_relationshipprop_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        FeatureRelationshipProp.feature_relationshipprop_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    pub_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        pub.Pub.pub_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)

    # Constraints
    __tablename__ = "feature_relationshipprop_pub"
    __table_args__ = (sqlalchemy.UniqueConstraint(feature_relationshipprop_id, pub_id,
                                                  name="feature_relationshipprop_pub_c1"),
                      sqlalchemy.Index("feature_relationshipprop_pub_idx1", feature_relationshipprop_id),
                      sqlalchemy.Index("feature_relationshipprop_pub_idx2", pub_id))

    # Relationships
    feature_relationshipprop = sqlalchemy.orm.relationship(
        FeatureRelationshipProp, foreign_keys=feature_relationshipprop_id,
        backref="feature_relationshipprop_pub_feature_relationshipprop")
    pub = sqlalchemy.orm.relationship(pub.Pub, foreign_keys=pub_id, backref="feature_relationshipprop_pub_pub")

    # Initialisation
    def __init__(self, feature_relationshipprop_id, pub_id, feature_relationshipprop_pub_id=None):
        for key, value in locals().items():
            if key != self:
                setattr(self, key, value)

    # Representation
    def __repr__(self):
        return "<sequence.FeatureRelationshipPropPub(feature_relationshipprop_pub_id={0}, " \
               "feature_relationshipprop_id={1}, pub_id={2})>".\
            format(self.feature_relationshipprop_pub_id, self.feature_relationshipprop_id, self.pub_id)


class Synonym(base.PublicBase):
    """Class for the CHADO 'synonym' table"""
    # Columns
    synonym_id = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=False, primary_key=True, autoincrement=True)
    name = sqlalchemy.Column(sqlalchemy.VARCHAR(255), nullable=False)
    type_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        cv.CvTerm.cvterm_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    synonym_sgml = sqlalchemy.Column(sqlalchemy.VARCHAR(255), nullable=False)

    # Constraints
    __tablename__ = "synonym"
    __table_args__ = (sqlalchemy.UniqueConstraint(name, type_id, name="synonym_c1"),
                      sqlalchemy.Index("synonym_idx1", type_id),
                      sqlalchemy.Index("synonym_idx2", sqlalchemy.func.lower(synonym_sgml)))

    # Relationships
    type = sqlalchemy.orm.relationship(cv.CvTerm, foreign_keys=type_id, backref="synonym_type")

    # Initialisation
    def __init__(self, name, type_id, synonym_sgml, synonym_id=None):
        for key, value in locals().items():
            if key != self:
                setattr(self, key, value)

    # Representation
    def __repr__(self):
        return "<sequence.Synonym(synonym_id={0}, name='{1}', type_id={2}, synonym_sgml='{3}')>".\
            format(self.synonym_id, self.name, self.type_id, self.synonym_sgml)


class FeatureSynonym(base.PublicBase):
    """Class for the CHADO 'feature_synonym' table"""
    # Columns
    feature_synonym_id = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=False, primary_key=True, autoincrement=True)
    synonym_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        Synonym.synonym_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    feature_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        Feature.feature_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    pub_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        pub.Pub.pub_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    is_current = sqlalchemy.Column(sqlalchemy.BOOLEAN, nullable=False, server_default="True")
    is_internal = sqlalchemy.Column(sqlalchemy.BOOLEAN, nullable=False, server_default="False")

    # Constraints
    __tablename__ = "feature_synonym"
    __table_args__ = (sqlalchemy.UniqueConstraint(synonym_id, feature_id, pub_id, name="feature_synonym_c1"),
                      sqlalchemy.Index("feature_synonym_idx1", synonym_id),
                      sqlalchemy.Index("feature_synonym_idx2", feature_id),
                      sqlalchemy.Index("feature_synonym_idx3", pub_id))

    # Relationships
    synonym = sqlalchemy.orm.relationship(Synonym, foreign_keys=synonym_id, backref="feature_synonym_synonym")
    feature = sqlalchemy.orm.relationship(Feature, foreign_keys=feature_id, backref="feature_synonym_feature")
    pub = sqlalchemy.orm.relationship(pub.Pub, foreign_keys=pub_id, backref="feature_synonym_pub")

    # Initialisation
    def __init__(self, synonym_id, feature_id, pub_id, is_current=True, is_internal=False, feature_synonym_id=None):
        for key, val in locals().items():
            if key != self:
                setattr(self, key, val)

    # Representation
    def __repr__(self):
        return "<sequence.FeatureSynonym(feature_synonym_id={0}, synonym_id={1}, feature_id={2}, pub_id={3}, " \
               "is_current={4}, is_internal={5})>".format(self.feature_synonym_id, self.synonym_id, self.feature_id,
                                                          self.pub_id, self.is_current, self.is_internal)


class FeatureLoc(base.PublicBase):
    """Class for the CHADO 'featureloc' table"""
    # Columns
    featureloc_id = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=False, primary_key=True, autoincrement=True)
    feature_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        Feature.feature_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    srcfeature_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        Feature.feature_id, onupdate="CASCADE", ondelete="SET NULL"), nullable=True)
    fmin = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=True)
    is_fmin_partial = sqlalchemy.Column(sqlalchemy.BOOLEAN, nullable=False, server_default="False")
    fmax = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=True)
    is_fmax_partial = sqlalchemy.Column(sqlalchemy.BOOLEAN, nullable=False, server_default="False")
    strand = sqlalchemy.Column(sqlalchemy.SMALLINT, nullable=True)
    phase = sqlalchemy.Column(sqlalchemy.INTEGER, nullable=True)
    residue_info = sqlalchemy.Column(sqlalchemy.TEXT, nullable=True)
    locgroup = sqlalchemy.Column(sqlalchemy.INTEGER, nullable=False, server_default="0")
    rank = sqlalchemy.Column(sqlalchemy.INTEGER, nullable=False, server_default="0")

    # Constraints
    __tablename__ = "featureloc"
    __table_args__ = (sqlalchemy.UniqueConstraint(feature_id, locgroup, rank, name="featureloc_c1"),
                      sqlalchemy.CheckConstraint("fmin <= fmax", name="featureloc_c2"),
                      sqlalchemy.Index("featureloc_idx1", feature_id),
                      sqlalchemy.Index("featureloc_idx2", srcfeature_id),
                      sqlalchemy.Index("featureloc_idx3", srcfeature_id, fmin, fmax))

    # Relationships
    feature = sqlalchemy.orm.relationship(Feature, foreign_keys=feature_id, backref="featureloc_feature")
    srcfeature = sqlalchemy.orm.relationship(Feature, foreign_keys=srcfeature_id, backref="featureloc_srcfeature")

    # Initialisation
    def __init__(self, feature_id, srcfeature_id, fmin=None, is_fmin_partial=False, fmax=None, is_fmax_partial=False,
                 strand=None, phase=None, residue_info=None, locgroup=0, rank=0, featureloc_id=None):
        for key, value in locals().items():
            if key != self:
                setattr(self, key, value)

    # Representation
    def __repr__(self):
        return "<sequence.FeatureLoc(featureloc_id={0}, feature_id={1}, srcfeature_id={2}, fmin={3}, " \
               "is_fmin_partial={4}, fmax={5}, is_fmax_partial={6}, strand={7}, phase={8}, residue_info='{9}', " \
               "locgroup={10}, rank={11})>".format(self.featureloc_id, self.feature_id, self.srcfeature_id, self.fmin,
                                                   self.is_fmin_partial, self.fmax, self.is_fmax_partial, self.strand,
                                                   self.phase, self.residue_info, self.locgroup, self.rank)


class FeatureLocPub(base.PublicBase):
    """Class for the CHADO 'featureloc_pub' table"""
    # Columns
    featureloc_pub_id = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=False, primary_key=True, autoincrement=True)
    featureloc_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        FeatureLoc.featureloc_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    pub_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        pub.Pub.pub_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)

    # Constraints
    __tablename__ = "featureloc_pub"
    __table_args__ = (sqlalchemy.UniqueConstraint(featureloc_id, pub_id, name="featureloc_pub_c1"),
                      sqlalchemy.Index("featureloc_pub_idx1", featureloc_id),
                      sqlalchemy.Index("featureloc_pub_idx2", pub_id))

    # Relationships
    featureloc = sqlalchemy.orm.relationship(FeatureLoc, foreign_keys=featureloc_id,
                                             backref="featureloc_pub_featureloc")
    pub = sqlalchemy.orm.relationship(pub.Pub, foreign_keys=pub_id, backref="featureloc_pub_pub")

    # Initialisation
    def __init__(self, featureloc_id, pub_id, featureloc_pub_id=None):
        for key, value in locals().items():
            if key != self:
                setattr(self, key, value)

    # Representation
    def __repr__(self):
        return "<sequence.FeatureLocPub(featureloc_pub_id={0}, featureloc_id={1}, pub_id={2})>".\
            format(self.featureloc_pub_id, self.featureloc_id, self.pub_id)


class FeatureProp(base.PublicBase):
    """Class for the CHADO 'featureprop' table"""
    # Columns
    featureprop_id = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=False, primary_key=True, autoincrement=True)
    feature_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        Feature.feature_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    type_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        cv.CvTerm.cvterm_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    value = sqlalchemy.Column(sqlalchemy.TEXT, nullable=True)
    rank = sqlalchemy.Column(sqlalchemy.INTEGER, nullable=False, server_default="0")

    # Constraints
    __tablename__ = "featureprop"
    __table_args__ = (sqlalchemy.UniqueConstraint(feature_id, type_id, rank, name="featureprop_c1"),
                      sqlalchemy.Index("featureprop_idx1", feature_id),
                      sqlalchemy.Index("featureprop_idx2", type_id))

    # Relationships
    feature = sqlalchemy.orm.relationship(Feature, foreign_keys=feature_id, backref="featureprop_feature")
    type = sqlalchemy.orm.relationship(cv.CvTerm, foreign_keys=type_id, backref="featureprop_type")

    # Initialisation
    def __init__(self, feature_id, type_id, value=None, rank=0, featureprop_id=None):
        for key, val in locals().items():
            if key != self:
                setattr(self, key, val)

    # Representation
    def __repr__(self):
        return "<sequence.FeatureProp(featureprop_id={0}, feature_id={1}, type_id={2}, value='{3}', rank={4})>". \
            format(self.featureprop_id, self.feature_id, self.type_id, self.value, self.rank)


class FeaturePropPub(base.PublicBase):
    """Class for the CHADO 'featureprop_pub' table"""
    # Columns
    featureprop_pub_id = sqlalchemy.Column(sqlalchemy.BIGINT, nullable=False, primary_key=True, autoincrement=True)
    featureprop_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        FeatureProp.featureprop_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)
    pub_id = sqlalchemy.Column(sqlalchemy.BIGINT, sqlalchemy.ForeignKey(
        pub.Pub.pub_id, onupdate="CASCADE", ondelete="CASCADE"), nullable=False)

    # Constraints
    __tablename__ = "featureprop_pub"
    __table_args__ = (sqlalchemy.UniqueConstraint(featureprop_id, pub_id, name="featureprop_pub_c1"),
                      sqlalchemy.Index("featureprop_pub_idx1", featureprop_id),
                      sqlalchemy.Index("featureprop_pub_idx2", pub_id))

    # Relationships
    featureprop = sqlalchemy.orm.relationship(FeatureProp, foreign_keys=featureprop_id,
                                              backref="featureprop_pub_featureprop")
    pub = sqlalchemy.orm.relationship(pub.Pub, foreign_keys=pub_id, backref="featureprop_pub_pub")

    # Initialisation
    def __init__(self, featureprop_id, pub_id, featureprop_pub_id=None):
        for key, value in locals().items():
            if key != self:
                setattr(self, key, value)

    # Representation
    def __repr__(self):
        return "<sequence.FeaturePropPub(featureprop_pub_id={0}, featureprop_id={1}, pub_id={2})>".\
            format(self.featureprop_pub_id, self.featureprop_id, self.pub_id)



import os
import unittest.mock
import urllib.error
import tempfile
import filecmp
import getpass
import sqlalchemy.engine
import sqlalchemy_utils
from .. import dbutils, utils


class TestConnection(unittest.TestCase):
    """Tests for database connections"""

    modules_dir = os.path.dirname(os.path.abspath(dbutils.__file__))
    data_dir = os.path.join(modules_dir, 'tests', 'data')

    def setUp(self):
        # Checks if the default connection file is available and reads in the parameters
        self.assertTrue(os.path.exists(os.path.abspath(dbutils.default_configuration_file())))
        self.connectionParameters = utils.parse_yaml(dbutils.default_configuration_file())
        self.uri = dbutils.generate_uri(self.connectionParameters)

    def tearDown(self):
        self.connectionParameters.clear()

    def test_factory_settings(self):
        # Tests if the settings in the default connection file are equivalent to the factory settings
        factory_settings = utils.parse_yaml(dbutils.factory_settings_configuration_file())
        self.assertEqual(self.connectionParameters, factory_settings)

    def test_connection_parameters(self):
        # Tests if the default connection file contains all required parameters
        self.assertIn("database", self.connectionParameters)
        self.assertIn("user", self.connectionParameters)
        self.assertIn("password", self.connectionParameters)
        self.assertIn("host", self.connectionParameters)
        self.assertIn("port", self.connectionParameters)

    @unittest.mock.patch('builtins.input')
    @unittest.mock.patch('getpass.getpass')
    def test_set_reset_parameters(self, mock_getpass, mock_input):
        # Tests if the default connection parameters can be changed
        self.assertIs(mock_input, input)
        self.assertIs(mock_getpass, getpass.getpass)
        mock_input.side_effect = ["myhost", 5555, "mydb", "myuser"]
        mock_getpass.return_value = "mypw"
        dbutils.set_default_parameters()
        default_parameters = utils.parse_yaml(dbutils.default_configuration_file())
        self.assertEqual(default_parameters["host"], "myhost")
        self.assertEqual(default_parameters["port"], "5555")
        self.assertEqual(default_parameters["database"], "mydb")
        self.assertEqual(default_parameters["user"], "myuser")
        self.assertEqual(default_parameters["password"], "mypw")
        dbutils.reset_default_parameters()
        default_parameters = utils.parse_yaml(dbutils.default_configuration_file())
        self.assertEqual(self.connectionParameters, default_parameters)

    def test_connection_uri(self):
        # Tests the correct creation of a database connection string in URI format
        uri = "postgresql://" \
              + self.connectionParameters["user"] + ":" \
              + self.connectionParameters["password"] + "@" \
              + self.connectionParameters["host"] + ":" \
              + self.connectionParameters["port"] + "/" \
              + self.connectionParameters["database"]
        self.assertEqual(self.uri, uri)

    def test_random_database_uri(self):
        # Tests that a function creates a URI of a database that does not exist
        uri = dbutils.random_database_uri(self.connectionParameters)
        self.assertFalse(dbutils.exists(uri))
        pos = self.uri.rfind("/")
        self.assertEqual(uri[:pos], self.uri[:pos])
        self.assertNotEqual(uri[pos:], self.uri[pos:])

    def test_connect(self):
        # Tests that a connection to the default database can be established and that queries can be executed
        conn = dbutils.open_connection(self.uri)
        self.assertFalse(conn.closed)
        result = conn.execute("SELECT 1 + 2").scalar()
        self.assertEqual(result, 3)
        dbutils.close_connection(conn)
        self.assertTrue(conn.closed)

    @unittest.mock.patch('builtins.input')
    @unittest.mock.patch('sqlalchemy_utils.drop_database')
    def test_drop_on_demand(self, mock_drop, mock_input):
        # Tests that a database is only dropped after confirmation by the user
        self.assertIs(mock_drop, sqlalchemy_utils.drop_database)
        self.assertIs(mock_input, input)

        mock_input.return_value = 'n'
        dbutils.drop_database(self.uri)
        mock_drop.assert_not_called()

        mock_drop.reset_mock()
        mock_input.return_value = 'y'
        dbutils.drop_database(self.uri)
        mock_drop.assert_called()

        mock_drop.reset_mock()
        dbutils.drop_database(self.uri, True)
        mock_drop.assert_called()

    def test_integration(self):
        # Test the basic functionality for creation and deletion of databases
        # NOTE: This test depends on an example SQL schema in the tests/data directory.
        # If the schema is changed, the test might fail.

        # Generate a database with a random name and check it exists
        uri = dbutils.random_database_uri(self.connectionParameters)
        dbutils.create_database(uri)
        self.assertTrue(dbutils.exists(uri))

        # Set up the database according to a test schema
        test_schema = os.path.join(self.data_dir, "dbutils_example_schema.sql")
        self.assertTrue(os.path.exists(test_schema))
        dbutils.setup_database(uri, test_schema)

        # Check if the database is correctly set up
        conn = dbutils.open_connection(uri)
        result_proxy = conn.execute("SELECT * FROM species ORDER BY legs ASC")
        result = result_proxy.fetchall()
        self.assertEqual(len(result), 4)
        self.assertEqual(result[0]["name"], "leech")
        result_proxy.close()
        dbutils.close_connection(conn)

        # Dump the database and check for the archive file
        archive_file = "tmp.dump"
        dbutils.dump_database(uri, archive_file)
        self.assertTrue(os.path.exists(archive_file))

        # Restore the database and check it exists
        dbutils.restore_database(uri, archive_file)
        self.assertTrue(dbutils.exists(uri))

        # Check if the database is still correctly set up
        conn = dbutils.open_connection(uri)
        result_proxy = conn.execute("SELECT name FROM species WHERE extinct = TRUE")
        result = result_proxy.fetchall()
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0]["name"], "diplodocus")
        result_proxy.close()

        # Export the entire table and check that the result is as expected
        result = dbutils.run_query(conn, "SELECT name, clade, legs FROM species ORDER BY legs ASC").fetchall()
        self.assertEqual(len(result), 4)
        self.assertEqual(len(result[0]), 3)
        self.assertEqual(result[0][0], "leech")
        dbutils.close_connection(conn)

        # Drop the database, remove the archive file and check that everything is gone
        dbutils.drop_database(uri, True)
        self.assertFalse(dbutils.exists(uri))
        os.remove(archive_file)
        self.assertFalse(os.path.exists(archive_file))

    @unittest.mock.patch('pychado.dbutils.print_query_result')
    @unittest.mock.patch('pychado.dbutils.run_query')
    @unittest.mock.patch('pychado.dbutils.close_connection')
    @unittest.mock.patch('pychado.dbutils.open_connection')
    def test_query_and_print(self, mock_open, mock_close, mock_query, mock_print):
        # Tests that the wrapper function for database queries is correctly run
        self.assertIs(mock_open, dbutils.open_connection)
        self.assertIs(mock_close, dbutils.close_connection)
        self.assertIs(mock_query, dbutils.run_query)
        self.assertIs(mock_print, dbutils.print_query_result)

        mock_connection_object = mock_open.return_value
        mock_result_object = mock_query.return_value

        dbutils.query_and_print("testuri", "testquery", "testfile", "json", True, ";")
        mock_open.assert_called_with("testuri")
        mock_query.assert_called_with(mock_connection_object, "testquery")
        mock_print.assert_called_with(mock_result_object, "testfile", "json", True, ";")
        self.assertIn(unittest.mock.call.close(), mock_result_object.mock_calls)
        mock_close.assert_called_with(mock_connection_object)

    @unittest.mock.patch('pychado.dbutils.print_query_result_csv')
    @unittest.mock.patch('pychado.dbutils.print_query_result_json')
    @unittest.mock.patch('pychado.utils.close')
    @unittest.mock.patch('pychado.utils.open_file_write')
    def test_print_query_result(self, mock_open, mock_close, mock_json, mock_csv):
        # Tests the function printing database queries to file
        self.assertIs(mock_open, utils.open_file_write)
        self.assertIs(mock_close, utils.close)
        self.assertIs(mock_json, dbutils.print_query_result_json)
        self.assertIs(mock_csv, dbutils.print_query_result_csv)

        result_object = unittest.mock.Mock(spec=sqlalchemy.engine.ResultProxy)
        mock_file_object = mock_open.return_value

        dbutils.print_query_result(result_object, "testfile", "json", True, ";")
        mock_open.assert_called_with("testfile")
        mock_json.assert_called_with(result_object, mock_file_object)
        mock_csv.assert_not_called()
        mock_close.assert_called_with(mock_file_object)

        mock_json.reset_mock()
        mock_csv.reset_mock()
        dbutils.print_query_result(result_object, "testfile", "csv", True, ";")
        mock_json.assert_not_called()
        mock_csv.assert_called_with(result_object, mock_file_object, True, ";")

    def test_print_json(self):
        # Tests the function exporting a query result to a JSON file
        result_object = unittest.mock.Mock(spec=sqlalchemy.engine.ResultProxy)
        result_object.configure_mock(**{"keys.return_value": ["key1", "key2"],
                                        "fetchone.side_effect": [["v1", "v2"], ["x1", "x2"], None]})
        filename = tempfile.mkstemp()[1]
        f = utils.open_file_write(filename)
        dbutils.print_query_result_json(result_object, f)
        utils.close(f)
        self.assertTrue(filecmp.cmp(filename, os.path.join(self.data_dir, "dbutils_test_result.json")))
        os.remove(filename)

    def test_print_csv(self):
        # Tests the function exporting a query result to a CSV file
        result_object = unittest.mock.Mock(spec=sqlalchemy.engine.ResultProxy)
        result_object.configure_mock(**{"keys.return_value": ["key1", "key2"],
                                        "fetchone.side_effect": [["v1", "v2"], ["x1", "x2"], None]})
        filename = tempfile.mkstemp()[1]
        f = utils.open_file_write(filename)
        dbutils.print_query_result_csv(result_object, f, True, ";")
        utils.close(f)
        self.assertTrue(filecmp.cmp(filename, os.path.join(self.data_dir, "dbutils_test_result.csv")))
        os.remove(filename)


class TestDownload(unittest.TestCase):
    """Tests for data download"""

    def test_default_schema_url(self):
        # Tests if the default schema is retrievable and has a valid address
        url = dbutils.default_schema_url()
        self.assertEqual(url[:4], "http")
        self.assertEqual(url[-3:], "sql")

    def test_download_schema(self):
        # Tests the download of a file from a given url
        url = dbutils.default_schema_url()
        utils.download_file(url)
        url = url + "_arbitraryString"
        with self.assertRaises(urllib.error.HTTPError):
            utils.download_file(url)
        url = "http://xyzxyz.xyzxyz"
        with self.assertRaises(urllib.error.URLError):
            utils.download_file(url)


if __name__ == '__main__':
    unittest.main(verbosity=2, buffer=True)



#!/usr/bin/env python3

import setuptools
import os

try:
    with open(os.path.dirname(os.path.abspath(__file__)) + "/README.md") as readme_file:
        long_description = readme_file.read()
except FileNotFoundError or FileExistsError:
    long_description = "Tools to access CHADO databases. For a detailed description visit the GitHub repository."

setuptools.setup(
    name="chado-tools",
    version="0.2.5",
    author="Christoph Puethe",
    author_email="path-help@sanger.ac.uk",
    description="Tools to access CHADO databases",
    long_description_content_type="text/markdown",
    long_description=long_description,
    url="https://github.com/sanger-pathogens/chado-tools/",
    packages=setuptools.find_packages(),
    package_data={
        "pychado": [
            "data/*.yml",
            "sql/*.sql"
        ],
        "pychado.tests": [
            "data/*"
        ]
    },
    entry_points={
        "console_scripts": [
            "chado = pychado.chado_tools:main",
        ]
    },
    test_suite="nose.collector",
    tests_require=[
        "nose >= 1.3"
    ],
    install_requires=[
        "sqlalchemy",
        "sqlalchemy-utils",
        "psycopg2-binary",
        "pyyaml",
        "pronto >= 0.11.0",
        "gffutils",
        "biopython"
    ],
    license="GPLv3",
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Development Status :: 3 - Alpha"
    ]
)
